//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// Decl.hpp --
//
// Author           : Aaron B. Moss
// Created On       : Thu May 9 10:00:00 2019
// Last Modified By : Aaron B. Moss
// Last Modified On : Thu May 9 10:00:00 2019
// Update Count     : 1
//

#pragma once

#include <string>              // for string, to_string
#include <unordered_map>
#include <vector>

#include "FunctionSpec.hpp"
#include "Fwd.hpp"             // for UniqueId
#include "LinkageSpec.hpp"
#include "Node.hpp"            // for ptr, readonly
#include "ParseNode.hpp"
#include "StorageClasses.hpp"
#include "Type.hpp"            // for Type, ptr<Type>
#include "Visitor.hpp"
#include "Parser/ParseNode.h"  // for DeclarationNode::Aggregate

namespace ast {

class Attribute;
class Expr;
class TypeDecl;

/// Base declaration class
class Decl : public ParseNode {
public:
	std::string name;
	Storage::Classes storage;
	Linkage::Spec linkage;
	UniqueId uniqueId = 0;
	bool extension = false;

	Decl( const CodeLocation& loc, const std::string& name, Storage::Classes storage, 
		Linkage::Spec linkage )
	: ParseNode( loc ), name( name ), storage( storage ), linkage( linkage ) {}

	Decl* set_extension( bool ex ) { extension = ex; return this; }

	/// Ensures this node has a unique ID
	void fixUniqueId();
	/// Get canonical declaration for unique ID
	static readonly<Decl> fromId( UniqueId id );

	virtual Decl* accept( Visitor& v ) override = 0;
private:
	virtual Decl* clone() const override = 0;
};

/// Typed declaration base class
class DeclWithType : public Decl {
public:
	/// Represents the type with all types and typedefs expanded.
	/// This field is generated by SymTab::Validate::Pass2
	std::string mangleName;
	/// Stores the scope level at which the variable was declared. 
	/// Used to access shadowed identifiers.
	int scopeLevel = 0;

	std::vector<ptr<Attribute>> attributes;
	Function::Specs funcSpec;
	ptr<Expr> asmName;
	bool isDeleted = false;

	DeclWithType( const CodeLocation& loc, const std::string& name, Storage::Classes storage, 
		Linkage::Spec linkage, std::vector<ptr<Attribute>>&& attrs, Function::Specs fs )
	: Decl( loc, name, storage, linkage ), mangleName(), attributes( std::move(attrs) ), 
		funcSpec(fs), asmName() {}
	
	std::string scopedMangleName() const { return mangleName + "_" + std::to_string(scopeLevel); }

	/// Get type of this declaration. May be generated by subclass
	virtual const Type* type() const = 0;
	/// Set type of this declaration. May be verified by subclass
	virtual void set_type(Type*) = 0;

	virtual DeclWithType* accept( Visitor& v ) override = 0;
private:
	virtual DeclWithType* clone() const override = 0;
};

/// Aggregate type declaration base class
class AggregateDecl : public Decl {
public:
	std::vector<ptr<Decl>> members;
	std::vector<ptr<TypeDecl>> parameters;
	std::vector<ptr<Attribute>> attributes;
	bool body = false;
	readonly<AggregateDecl> parent = {};

	AggregateDecl( const CodeLocation& loc, const std::string& name, 
		std::vector<ptr<Attribute>>&& attrs = {}, Linkage::Spec linkage = Linkage::Cforall )
	: Decl( loc, name, Storage::Classes{}, linkage ), members(), parameters(), 
	  attributes( std::move(attrs) ) {}
	
	AggregateDecl* set_body( bool b ) { body = b; return this; }

protected:
	/// Produces a name for the kind of aggregate
	virtual std::string typeString() const = 0;
};

/// struct declaration `struct Foo { ... };`
class StructDecl final : public AggregateDecl {
public:
	DeclarationNode::Aggregate kind;

	StructDecl( const CodeLocation& loc, const std::string& name, 
		DeclarationNode::Aggregate kind = DeclarationNode::Struct, 
		std::vector<ptr<Attribute>>&& attrs = {}, Linkage::Spec linkage = Linkage::Cforall )
	: AggregateDecl( loc, name, std::move(attrs), linkage ), kind( kind ) {}

	bool is_coroutine() { return kind == DeclarationNode::Coroutine; }
	bool is_monitor() { return kind == DeclarationNode::Monitor; }
	bool is_thread() { return kind == DeclarationNode::Thread; }

	Decl* accept( Visitor& v ) override { return v.visit( this ); }
private:
	StructDecl* clone() const override { return new StructDecl{ *this }; }

	std::string typeString() const override { return "struct"; }
};

/// union declaration `union Foo { ... };`
class UnionDecl final : public AggregateDecl {
public:
	UnionDecl( const CodeLocation& loc, const std::string& name, 
		std::vector<ptr<Attribute>>&& attrs = {}, Linkage::Spec linkage = Linkage::Cforall )
	: AggregateDecl( loc, name, std::move(attrs), linkage ) {}

	Decl* accept( Visitor& v ) override { return v.visit( this ); }
private:
	UnionDecl* clone() const override { return new UnionDecl{ *this }; }

	std::string typeString() const override { return "union"; }
};

/// enum declaration `enum Foo { ... };`
class EnumDecl final : public AggregateDecl {
public:
	EnumDecl( const CodeLocation& loc, const std::string& name, 
		std::vector<ptr<Attribute>>&& attrs = {}, Linkage::Spec linkage = Linkage::Cforall )
	: AggregateDecl( loc, name, std::move(attrs), linkage ), enumValues() {}

	/// gets the integer value for this enumerator, returning true iff value found
	bool valueOf( Decl* enumerator, long long& value ) const;

	Decl* accept( Visitor& v ) override { return v.visit( this ); }
private:
	EnumDecl* clone() const override { return new EnumDecl{ *this }; }

	std::string typeString() const override { return "enum"; }

	/// Map from names to enumerator values; kept private for lazy initialization
	mutable std::unordered_map< std::string, long long > enumValues;
};

/// trait declaration `trait Foo( ... ) { ... };`
class TraitDecl final : public AggregateDecl {
public:
	TraitDecl( const CodeLocation& loc, const std::string& name, 
		std::vector<ptr<Attribute>>&& attrs = {}, Linkage::Spec linkage = Linkage::Cforall )
	: AggregateDecl( loc, name, std::move(attrs), linkage ) {}

	Decl* accept( Visitor& v ) override { return v.visit( this ); }
private:
	TraitDecl* clone() const override { return new TraitDecl{ *this }; }

	std::string typeString() const override { return "trait"; }
};

}

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
